# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import shlex
import stat
import subprocess
import tempfile
import time
from copy import deepcopy
from pathlib import Path
from typing import Any, Tuple, Optional
import shutil
import psutil
from loguru import logger

from msserviceprofiler.modelevalstate.config.config import (MindieConfig, MODEL_EVAL_STATE_CONFIG_PATH,
                                                            modelevalstate_config_path, CUSTOM_OUTPUT, custom_output)
from msserviceprofiler.modelevalstate.config.config import OptimizerConfigField
from msserviceprofiler.modelevalstate.optimizer.utils import (backup, kill_process, kill_children,
                                                              remove_file, close_file_fp)
from msserviceprofiler.msguard.security import open_s


class Simulator:
    from msserviceprofiler.modelevalstate.config.custom_command import MindieCommand

    def __init__(self, mindie_config: MindieConfig, bak_path: Optional[Path] = None):
        self.mindie_config = mindie_config
        logger.info(f"config path {self.mindie_config.config_path}", )
        if not self.mindie_config.config_path.exists():
            raise FileNotFoundError(self.mindie_config.config_path)
        with open_s(self.mindie_config.config_path, "r") as f:
            data = json.load(f)
        self.default_config = data
        logger.info(f"config bak path {self.mindie_config.config_bak_path}", )
        if self.mindie_config.config_bak_path.exists():
            self.mindie_config.config_bak_path.unlink()
        flags = os.O_WRONLY | os.O_CREAT
        modes = stat.S_IWUSR | stat.S_IRUSR
        with os.fdopen(os.open(self.mindie_config.config_bak_path, flags, modes), 'w') as fout:
            json.dump(self.default_config, fout, indent=4)
        self.mindie_log = None
        self.mindie_log_offset = 0
        self.bak_path = bak_path
        self.mindie_log_fp = None
        self.process = None
        self.command = self.MindieCommand(self.mindie_config.command).command

    @staticmethod
    def is_int(x):
        try:
            int(x)
            return True
        except ValueError:
            return False
        
    @staticmethod
    def prepare_before_start_server():
        pkill_path = shutil.which("pkill")
        if pkill_path is not None:
            subprocess.run([pkill_path, "-9", "mindie"])
        else:
            logger.error("pkill not found in path")

    @staticmethod
    def set_config_for_dict(origin_config, cur_key, next_key, next_level, value):
        if cur_key in origin_config:
            Simulator.set_config(origin_config[cur_key], next_level, value)
        elif Simulator.is_int(cur_key):
            raise KeyError(f"data: {origin_config}, key: {cur_key}")
        elif Simulator.is_int(next_key):
            origin_config[cur_key] = []
            Simulator.set_config(origin_config[cur_key], next_level, value)
        else:
            origin_config[cur_key] = {}
            Simulator.set_config(origin_config[cur_key], next_level, value)

    @staticmethod
    def set_config_for_list(origin_config, cur_key, next_key, next_level, value):
        if len(origin_config) > int(cur_key):
            Simulator.set_config(origin_config[int(cur_key)], next_level, value)
        elif len(origin_config) == int(cur_key) and Simulator.is_int(next_key):
            origin_config.append([])
            Simulator.set_config(origin_config[int(cur_key)], next_level, value)
        elif len(origin_config) == int(cur_key) and not Simulator.is_int(next_key):
            origin_config.append({})
            Simulator.set_config(origin_config[int(cur_key)], next_level, value)
        else:
            raise IndexError(f"data: {origin_config}, index: {cur_key}")

    @staticmethod
    def set_config(origin_config, key: str, value: Any):
        next_level = None
        if "." in key:
            _f_index = key.index(".")
            _cur_key, next_level = key[:_f_index], key[_f_index + 1:]
        else:
            _cur_key = key
        if next_level is None:
            if isinstance(origin_config, dict):
                origin_config[_cur_key] = value
            elif isinstance(origin_config, list):
                if len(origin_config) > int(_cur_key):
                    origin_config[int(_cur_key)] = value
                else:
                    origin_config.append(value)
            return
        if "." in next_level:
            _next_index = next_level.index(".")
            _next_key, _next_next_level = next_level[:_next_index], next_level[_next_index + 1:]
        elif next_level:
            _next_key = next_level
            _next_next_level = None
        else:
            _next_key = None
        if isinstance(origin_config, dict):
            Simulator.set_config_for_dict(origin_config, _cur_key, _next_key, next_level, value)
        elif isinstance(origin_config, list):
            Simulator.set_config_for_list(origin_config, _cur_key, _next_key, next_level, value)
        else:
            raise ValueError(f"Not Support type {type(origin_config)}")

    def backup(self, del_log=True):
        backup(self.mindie_config.config_path, self.bak_path, self.__class__.__name__)
        if not del_log and self.mindie_log:
            backup(self.mindie_log, self.bak_path, self.__class__.__name__)

    def update_config(self, params: Tuple[OptimizerConfigField]):
        # 将params值更新到新的config中
        new_config = deepcopy(self.default_config)
        for p in params:
            if not p.config_position.startswith("BackendConfig"):
                continue
            Simulator.set_config(new_config, p.config_position, p.value)

        # 将新的config写入到config文件中
        logger.debug(f"new config {new_config}")
        flags = os.O_WRONLY | os.O_CREAT
        modes = stat.S_IWUSR | stat.S_IRUSR
        if self.mindie_config.config_path.exists():
            self.mindie_config.config_path.unlink()
        with os.fdopen(os.open(self.mindie_config.config_path, flags, modes), "w") as fout:
            json.dump(new_config, fout, indent=4)

    def check_env(self):
        logger.info("check env")
        _residual_process = []
        _all_process_name = self.mindie_config.process_name.split(",")
        for proc in psutil.process_iter(["pid", "name"]):
            if not hasattr(proc, "info"):
                continue
            _proc_flag = []
            for p in _all_process_name:
                if p not in proc.info["name"]:
                    _proc_flag.append(True)
                else:
                    _proc_flag.append(False)
            if all(_proc_flag):
                continue
            _residual_process.append(proc)
        if _residual_process:
            logger.info("kill residual_process")
            for _p_name in _all_process_name:
                try:
                    kill_process(_p_name)
                except Exception as e:
                    logger.error(f"Failed to kill process. {e}")
        time.sleep(1)

    def check_success(self, print_log=False):
        with open_s(self.mindie_log, "r") as f:
            try:
                f.seek(self.mindie_log_offset)
                output = f.read()
                self.mindie_log_offset = f.tell()
            except Exception as e:
                logger.info(f"Failed in read mindie log. error: {e}")
        if output:
            if print_log:
                logger.info(f"simulate out: \n{output}")
            if "Daemon start success!" in output:
                return True
        if self.process.poll() is not None:
            raise subprocess.SubprocessError(
                f"Failed in run mindie. return code: {self.process.returncode}. "
                f"Please check the service log or console output.")
        return False

    def start_server(self, run_params: Tuple[OptimizerConfigField]):
        self.prepare_before_start_server()
        self.mindie_log_fp, self.mindie_log = tempfile.mkstemp(prefix="modelevalstate_mindie")
        self.mindie_log_offset = 0
        if self.mindie_config.work_path:
            cwd = self.mindie_config.work_path
        else:
            cwd = os.getcwd()
        for k in run_params:
            if k.config_position == "env":
                os.environ[k.name] = str(k.value)
                _var_name = f"${k.name}"
                if _var_name in self.command:
                    _i = self.command.index(_var_name)
                    self.command[_i] = str(k.value)
        if MODEL_EVAL_STATE_CONFIG_PATH not in os.environ:
            os.environ[MODEL_EVAL_STATE_CONFIG_PATH] = str(modelevalstate_config_path)
        if CUSTOM_OUTPUT not in os.environ:
            os.environ[CUSTOM_OUTPUT] = str(custom_output)
        logger.debug(f"env {os.environ}")
        logger.info(f"run cmd: {self.command}, log path: {self.mindie_log}")
        self.process = subprocess.Popen(self.command, stdout=self.mindie_log_fp, stderr=subprocess.STDOUT, 
                                        env=os.environ, text=True, cwd=cwd)

    def run(self, run_params: Tuple[OptimizerConfigField]):
        logger.info(f'start run in simulator. run params: {run_params}')
        # 根据params 修改配置文件
        self.update_config(run_params)
        # 启动mindie仿真
        try:
            self.check_env()
        except Exception as e:
            logger.error(f"Failed to check env. {e}")
        self.start_server(run_params)

    def stop(self, del_log=True):
        logger.info("Stop simulator process")
        if self.bak_path:
            self.backup()
        close_file_fp(self.mindie_log_fp)
        if del_log:
            remove_file(self.mindie_log)
        self.mindie_log_offset = 0
        if not self.process:
            return
        _process_state = self.process.poll()
        if _process_state is not None:
            logger.info(f"mindie already. exit_code: {_process_state}")
            return
        try:
            children = psutil.Process(self.process.pid).children(recursive=True)
            self.process.kill()
            try:
                self.process.wait(10)
            except subprocess.TimeoutExpired:
                self.process.send_signal(9)
            if self.process.poll() is not None:
                logger.info(f"The {self.process.pid} process has been shut down.")
            else:
                logger.error(f"The {self.process.pid} process shutdown failed.")
            kill_children(children)
            kill_process(self.mindie_config.process_name)
            remove_file(self.mindie_config.config_path)
            flags = os.O_WRONLY | os.O_CREAT
            modes = stat.S_IWUSR | stat.S_IRUSR
            with os.fdopen(os.open(self.mindie_config.config_path, flags, modes), "w") as fout:
                json.dump(self.default_config, fout)
        except Exception as e:
            logger.error(f"Failed to stop simulator process. {e}")


class VllmSimulator(Simulator):
    from msserviceprofiler.modelevalstate.config.custom_command import VllmCommand

    def __init__(self, vllm_config: MindieConfig, bak_path: Optional[Path] = None):
        try:
            super().__init__(vllm_config, bak_path)
        except Exception as e:
            logger.info('VllmSimulator init failed')
        self.vllm_config = vllm_config
        self.mindie_log = None
        self.mindie_log_offset = 0
        self.bak_path = bak_path
        self.mindie_log_fp = None
        self.process = None
        self.command = self.VllmCommand(self.vllm_config.vllm_command).command

    @staticmethod
    def prepare_before_start_server():
        pkill_path = shutil.which("pkill")
        if pkill_path is not None:
            subprocess.run([pkill_path, "-15", "vllm"])
        else:
            logger.error("pkill not found in path")

    def run(self, run_params: Tuple[OptimizerConfigField]):
        logger.info(f'start run in simulator. run params: {run_params}')
        # 启动mindie仿真
        try:
            self.check_env()
        except Exception as e:
            logger.error(f"Failed to check env. {e}")
        self.start_server(run_params)

    def check_success(self, print_log=False):
        with open_s(self.mindie_log, "r") as f:
            try:
                f.seek(self.mindie_log_offset)
                output = f.read()
                self.mindie_log_offset = f.tell()
            except Exception as e:
                logger.info(f"Failed in read vllm log. error: {e}")
        if output:
            if print_log:
                logger.info(f"simulate out: \n{output}")
            if "Application startup complete." in output:
                return True
        if self.process.poll() is not None:
            raise subprocess.SubprocessError(
                f"Failed in run vllm. return code: {self.process.returncode}. "
                f"Please check the service log or console output.")
        return False

    def stop(self, del_log=True):
        logger.info("Stop simulator process")
        if self.bak_path:
            self.backup()
        close_file_fp(self.mindie_log_fp)
        if del_log:
            remove_file(self.mindie_log)
        self.mindie_log_offset = 0
        if not self.process:
            return
        _process_state = self.process.poll()
        if _process_state is not None:
            logger.info(f"vllm already. exit_code: {_process_state}")
            return
        try:
            children = psutil.Process(self.process.pid).children(recursive=True)
            self.process.kill()
            try:
                self.process.wait(10)
            except subprocess.TimeoutExpired:
                self.process.send_signal(9)
            if self.process.poll() is not None:
                logger.info(f"The {self.process.pid} process has been shut down.")
            else:
                logger.error(f"The {self.process.pid} process shutdown failed.")
            kill_children(children)
            kill_process(self.mindie_config.process_name)
        except Exception as e:
            logger.error(f"Failed to stop simulator process. {e}")


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass

from .base import BaseConstraint
from .constraint_builder import Path, where


@dataclass(frozen=True)
class Rule:
    input_file_read: BaseConstraint = where(
        os.getuid() == 0,
        Path.is_file(),
        Path.is_file() & ~Path.has_soft_link() &
        Path.is_readable() & ~Path.is_writable_to_group_or_others() &
        Path.is_consistent_to_current_user() & Path.is_size_reasonable(),
        description="current user is root"
    )

    input_file_exec: BaseConstraint = where(
        os.getuid() == 0,
        Path.is_file(),
        Path.is_file() & ~Path.has_soft_link() &
        Path.is_executable() & ~Path.is_writable_to_group_or_others() &
        Path.is_consistent_to_current_user() & Path.is_size_reasonable(),
        description="current user is root"
    )

    input_dir_traverse: BaseConstraint = where(
        os.getuid() == 0,
        Path.is_dir(),
        Path.is_dir() & ~Path.has_soft_link() & Path.is_readable() &
        Path.is_executable() & ~Path.is_writable_to_group_or_others() &
        Path.is_consistent_to_current_user(),
        description="current user is root"
    )

    output_path_create: BaseConstraint = ~Path.exists() & ~Path.is_name_too_long()
    output_path_overwrite: BaseConstraint = where(
        os.getuid() == 0,
        Path.is_file(),
        Path.is_file() & ~Path.has_soft_link() &
        Path.is_writable() & ~Path.is_writable_to_group_or_others() &
        Path.is_consistent_to_current_user(),
        description="current user is root"
    )
    output_path_write: BaseConstraint = where(
        Path.exists(),
        output_path_overwrite,
        ~Path.is_name_too_long() & Path.has_writable_parent_dir()
    )
    output_path_append = output_path_overwrite

    output_dir: BaseConstraint = where(
        os.getuid() == 0,
        Path.is_dir(),
        Path.is_dir() & ~Path.has_soft_link() &
        Path.is_writable() & ~Path.is_writable_to_group_or_others() &
        Path.is_consistent_to_current_user(),
        description="current user is root"
    )




# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.
import os
import shutil
from loguru import logger
from pydantic import BaseModel
from msserviceprofiler.msguard import Rule
 
 
class BenchmarkCommandConfig(BaseModel):
    dataset_path: str = ""
    dataset_type: str = ""
    model_name: str = ""
    model_path: str = ""
    test_type: str = "client"
    max_output_len: str = ""
    http: str = "http://127.0.0.1:1025"
    management_http: str = "http://127.0.0.1:1026"
    warmup_size: str = "1"
    tokenizer: str = "True"
 
 
class BenchmarkCommand:
    def __init__(self, benchmark_command_config: BenchmarkCommandConfig):
        self.process = shutil.which("benchmark")
        self.benchmark_command_config = benchmark_command_config
 
    @property
    def command(self):
        if not Rule.input_file_read.is_satisfied_by(self.benchmark_command_config.dataset_path):
            logger.error("the file of dataset_path is not safe, please check")
            return None
        
        return [self.process,
                "--DatasetPath", self.benchmark_command_config.dataset_path,
                "--DatasetType", self.benchmark_command_config.dataset_type,
                "--ModelName", self.benchmark_command_config.model_name,
                "--ModelPath", self.benchmark_command_config.model_path,
                "--TestType", self.benchmark_command_config.test_type,
                "--MaxOutputLen", self.benchmark_command_config.max_output_len,
                "--Http", self.benchmark_command_config.http,
                "--ManagementHttp", self.benchmark_command_config.management_http,
                "--Concurrency", "$CONCURRENCY",
                "--RequestRate", "$REQUESTRATE",
                "--WarmupSize", self.benchmark_command_config.warmup_size,
                "--Tokenizer", self.benchmark_command_config.tokenizer]
 
 
class VllmBenchmarkCommandConfig(BaseModel):
    backend: str = "vllm"
    host: str = "127.0.0.1"
    port: str = "6379"
    model: str = ""
    served_model_name: str = ""
    dataset_name: str = ""
    dataset_path: str = ""
    num_prompts: str = ""
 
 
class VllmBenchmarkCommand:
    def __init__(self, benchmark_command_config: VllmBenchmarkCommandConfig):
        self.benchmark_command_config = benchmark_command_config
 
    @property
    def command(self):
        if not Rule.input_file_read.is_satisfied_by(self.benchmark_command_config.dataset_path):
            logger.error("the file of dataset_path is not safe, please check")
            return None
        return ["python", self.benchmark_command_config.serving,
                "--backend", self.benchmark_command_config.backend,
                "--host", self.benchmark_command_config.host,
                "--port", self.benchmark_command_config.port,
                "--model", self.benchmark_command_config.model,
                "--served-model-name", self.benchmark_command_config.served_model_name,
                "--dataset-name", self.benchmark_command_config.dataset_name,
                "--dataset-path", self.benchmark_command_config.dataset_path,
                "--num-prompts", self.benchmark_command_config.num_prompts,
                "--max-concurrency", "$MAXCONCURRENCY",
                "--request-rate", "$REQUESTRATE",
                "--result-dir", "$MODEL_EVAL_STATE_VLLM_CUSTOM_OUTPUT",
                "--save-result"]
 
 
class MindieCommandConfig(BaseModel):
    pass
 
 
class MindieCommand:
    def __init__(self, command_config: MindieCommandConfig):
        self.command_config = command_config
 
    @property
    def command(self):
        mindie_service_default_path: str = "/usr/local/Ascend/mindie/latest/mindie-service"
        mindie_service_path: str = os.getenv("MIES_INSTALL_PATH", mindie_service_default_path)
        mindie_command_path: str = os.path.join(mindie_service_path, "bin", "mindieservice_daemon")
        return [mindie_command_path]
 
 
class VllmCommandConfig(BaseModel):
    host: str = "127.0.0.1"
    port: str = "6379"
    model: str = ""
    served_model_name: str = ""
 
 
class VllmCommand:
    def __init__(self, command_config: VllmCommandConfig):
        self.process = shutil.which("vllm")
        self.command_config = command_config
 
    @property
    def command(self):
        return [self.process, "serve",
                self.command_config.model,
                "--served-model-name", self.command_config.served_model_name,
                "--host", self.command_config.host,
                "--port", self.command_config.port,
                "--max-num-batched-tokens", "$MAX_NUM_BATCHED_TOKENS",
                "--max-num-seqs", "$MAX_NUM_SEQS"]


# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import traceback
import importlib
import inspect
import functools
from collections import namedtuple
from abc import ABC, abstractmethod
from typing import Union, Tuple, List, Optional, Callable, Dict, Any
from packaging.version import Version
from .utils import logger

# 全局注册表存储所有hookers
HOOK_REGISTRY = []


def import_object_from_string(import_path: str, module_path: str) -> Any:
    """
    根据点分隔的路径导入对象（模块、类、函数等）
    支持多级嵌套属性，如 "module.submodule.ClassName.method_name"
    """
    try:
        module = importlib.import_module(import_path)
    except ImportError as e:
        logger.error(f"Module import failed for %s: %s", module_path, e)
        return None

    current = module
    for part in module_path.split("."):
        current = getattr(current, part, None)
        if current is None:
            logger.error("Module {current} doesn't has attribute {part}")
            return None
    return current


class HookHelper:
    """辅助类，用于函数替换操作"""
    def __init__(self, ori_function_define, new_function):
        self.new_function = new_function
        self.ori_function = None
        self.location = None
        self.attr_name = None

        if ori_function_define is None:
            return

        if inspect.isfunction(ori_function_define) or inspect.ismethod(ori_function_define):
            self.ori_function = ori_function_define
            self.location, self.attr_name = self.get_location(self.ori_function)
        elif callable(ori_function_define):
            self.ori_function = ori_function_define.__call__
            self.location, self.attr_name = self.get_location(self.ori_function)

        if not all((self.ori_function, self.location, self.attr_name, self.new_function)):
            warn_msg = f"{ori_function_define} replace failed."
            logger.error(warn_msg)
            raise ValueError(warn_msg)

    @staticmethod
    def get_location(function_ins):
        if not hasattr(function_ins, "__module__"):
            warning_msg = f"function {str(function_ins)} has no __module__."
            logger.error(warning_msg)
            raise ValueError(warning_msg)

        module = importlib.import_module(function_ins.__module__)
        qualified_name = function_ins.__qualname__.split(".")
        classes = qualified_name[:-1]
        attr_name = qualified_name[-1]
        location = module

        for class_name in classes:
            location = getattr(location, class_name, None)
            if location is None:
                break

        if location is None:
            warning_msg = f"{'.'.join(classes)} does not exist"
            logger.error(warning_msg)
            raise ValueError(warning_msg)

        return location, attr_name

    def replace(self):
        if all((self.ori_function, self.location, self.attr_name, self.new_function)):
            setattr(self.location, self.attr_name, self._get_method(self.new_function))

    def recover(self):
        if all((self.ori_function, self.location, self.attr_name, self.new_function)):
            setattr(self.location, self.attr_name, self._get_method(self.ori_function))

    def _get_method(self, func):
        if inspect.isclass(self.location):
            try:
                func_cls_name = inspect.getattr_static(self.location, self.attr_name).__class__.__name__
                if func_cls_name in ("staticmethod", "classmethod"):
                    return staticmethod(func)
            except AttributeError:
                pass
        return func


def get_parents_name(ori_func, index=1):
    """获取调用栈中的父级函数名"""
    gen = traceback.walk_stack(None)
    try:
        for _ in range(index + 1):
            f = next(gen)
        return f[0].f_code.co_name
    except StopIteration:
        return None


class VLLMHookerBase(ABC):
    """Hooker基类，提供核心功能"""
    def __init__(self):
        self.hooks = []
        self.vllm_version = (None, None)  # (min_version, max_version)

    @abstractmethod
    def init(self):
        """初始化hook点，子类必须实现"""
        pass

    def replace_func(self, ori_func, pname, profiler_func):
        """创建替换函数"""
        @functools.wraps(ori_func)
        def wrapper(*args, **kwargs):
            if pname is not None and self.get_parents_name(ori_func) != pname:
                logger.debug(f"calling {ori_func}")
                return ori_func(*args, **kwargs)
            logger.debug(f"calling profiler_func for {ori_func}")
            return profiler_func(*args, **kwargs)
        return wrapper

    def do_hook(self, hook_points, profiler_func_maker, pname=None):
        """执行实际的hook操作"""
        for ori_func in hook_points:
            profiler_func = profiler_func_maker(ori_func)
            HookHelper(ori_func, self.replace_func(ori_func, pname, profiler_func)).replace()
            logger.debug(f"replacing {ori_func}")

    def support_version(self, version):
        """检查当前版本是否支持"""
        min_version = self.vllm_version[0]
        max_version = self.vllm_version[1]

        if min_version is not None and Version(min_version) > Version(version):
            return False
        if max_version is not None and Version(max_version) < Version(version):
            return False
        return True

    def register(self):
        """注册hooker到全局注册表"""
        HOOK_REGISTRY.append(self)


def vllm_hook(
    hook_points: Union[Tuple[str, str], List[Tuple[str, str]]],
    min_version: Optional[str] = None,
    max_version: Optional[str] = None,
    caller_filter: Optional[str] = None
) -> Callable:
    """
    装饰器工厂函数，用于简化hooker创建
    :param hook_points: hook点列表，格式为(模块名, 属性路径)
    :param min_version: 支持的最小版本
    :param max_version: 支持的最大版本
    :param caller_filter: 调用者过滤条件
    """
    def decorator(hook_func):
        class AutoHooker(VLLMHookerBase):
            vllm_version = (min_version, max_version)

            def init(self):
                hook_list = [hook_points] if isinstance(hook_points, tuple) else hook_points
                points = [import_object_from_string(import_path, func_path) for import_path, func_path in hook_list]
                self.do_hook(
                    hook_points=points,
                    profiler_func_maker=lambda ori_func: lambda *args, **kwargs: hook_func(ori_func, *args, **kwargs),
                    pname=caller_filter,
                )
        hooker = AutoHooker()
        hooker.register()
        return hook_func

    return decorator


def apply_hooks(version: str = None):
    """应用所有注册的hookers"""
    if version is None:
        import vllm
        version = vllm.__version__
    for hooker in HOOK_REGISTRY:
        if hooker.support_version(version):
            try:
                hooker.init()
                logger.debug(f"Applied hooker: {hooker.__class__.__name__}")
            except Exception as e:
                logger.error(f"Failed to apply hooker: {str(e)}")
        else:
            logger.debug(f"Skipping hooker: {hooker.__class__.__name__} for version not matched")



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from functools import wraps

from ..constraints import BaseConstraint, InvalidParameterError
from .decorator_usage import validate_params


def validate_args(constraint: BaseConstraint, silent=True):
    """Usage: parser.add_argument("xx", type=validate_args(xxx))
    pass the `wrapper` to argparse, pass `arg` to `arg_check`and catch the error msg
    """
    @validate_params({"arg": constraint}) # any checks
    def arg_check(arg: str):
        return arg

    @wraps(arg_check)
    def wrapper(arg):
        try:
            return arg_check(arg)
        except InvalidParameterError as e:
            if silent:
                raise argparse.ArgumentTypeError(str(e))
            raise

    return wrapper



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

# limitations under the License.

from .path import (
    IsNameTooLong, IsFile, Exists, IsDir, HasSoftLink, IsReadable, 
    IsWritable, IsExecutable, IsWritableToGroupOrOthers, HasWritableParentDir,
    IsConsistentToCurrentUser, IsSizeReasonable
)
from ..constraints.base import BaseConstraint
from ..constraints.logic import FunctionConstraint, IfElseConstraint
from ..utils.constants import TYPE_ERROR_MSG


class PathConstraintBuilder:
    """
    Constraint builder for IDE support.

    Available constraints:
        is_file, file_exists, is_dir, has_soft_link, is_readable, is_writable,
        is_executable, is_not_writable_to_group_or_others, is_consistent_to_current_user,
        is_size_reasonable
    """
    @staticmethod
    def is_name_too_long():
        return IsNameTooLong()
        
    @staticmethod
    def is_file(): 
        return IsFile()
    
    @staticmethod
    def exists(): 
        return Exists()
    
    @staticmethod
    def is_dir(): 
        return IsDir()
    
    @staticmethod
    def has_soft_link(): 
        return HasSoftLink()
    
    @staticmethod
    def is_readable(): 
        return IsReadable()
    
    @staticmethod
    def is_writable(): 
        return IsWritable()
    
    @staticmethod
    def has_writable_parent_dir():
        return HasWritableParentDir()
    
    @staticmethod
    def is_executable(): 
        return IsExecutable()
    
    @staticmethod
    def is_writable_to_group_or_others():
        return IsWritableToGroupOrOthers()
    
    @staticmethod
    def is_consistent_to_current_user():
        return IsConsistentToCurrentUser()
    
    @staticmethod
    def is_size_reasonable(*, size_limit=None, require_confirm=True):
        return IsSizeReasonable(size_limit=size_limit, require_confirm=require_confirm)


Path = PathConstraintBuilder


def make_constraint(constraint, description=None):
    if isinstance(constraint, BaseConstraint):
        return constraint

    if callable(constraint):
        return FunctionConstraint(constraint, description)

    raise TypeError(
        TYPE_ERROR_MSG.format('constraint', 'BaseContraint or Callable', type(constraint).__name__)
    )


def where(condition, if_constraint, else_constraint, *, description=None):
    return IfElseConstraint(
        condition,
        make_constraint(if_constraint),
        make_constraint(else_constraint),
        description=description
    )



# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2025-2025. All rights reserved.
import os
import time
from enum import Enum
from pathlib import Path

import msserviceprofiler.modelevalstate

RUN_TIME = time.strftime("%Y%m%d%H%M%S", time.localtime())
INSTALL_PATH = Path(msserviceprofiler.modelevalstate.__path__[0])
RUN_PATH = Path(os.getcwd())
MODEL_EVAL_STATE_CONFIG_PATH = "MODEL_EVAL_STATE_CONFIG_PATH"
modelevalstate_config_path = os.getenv(MODEL_EVAL_STATE_CONFIG_PATH) or os.getenv(MODEL_EVAL_STATE_CONFIG_PATH.lower())
if not modelevalstate_config_path:
    modelevalstate_config_path = RUN_PATH.joinpath("config.json")
modelevalstate_config_path = Path(modelevalstate_config_path).absolute().resolve()

CUSTOM_OUTPUT = "MODEL_EVAL_STATE_CUSTOM_OUTPUT"
custom_output = os.getenv(CUSTOM_OUTPUT) or os.getenv(CUSTOM_OUTPUT.lower())
if custom_output:
    custom_output = Path(custom_output).resolve()
else:
    custom_output = RUN_PATH
FOLDER_LIMIT_SIZE = 1024 * 1024 * 1024


class DeployPolicy(Enum):
    single = "single"
    multiple = "multiple"


class BenchMarkPolicy(Enum):
    benchmark = "benchmark"
    profiler_benchmark = "profiler_benchmark"
    vllm_benchmark = "vllm_benchmark"


class AnalyzeTool(Enum):
    default = "default"
    profiler = "profiler"
    vllm_benchmark = "vllm"


class ServiceType(Enum):
    master = "master"
    slave = "slave"


# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from .utils import logger, set_log_level
from .module_hook import apply_hooks

if os.environ.get('VLLM_USE_V1', '0') == "0":
    from .vllm_v0 import batch_hookers, kvcache_hookers, model_hookers, request_hookers
    apply_hooks()  # 应用所有hookers
else:
    logger.error("vLLM V1 interface is not supported yet")



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import os
import re
import shlex
import subprocess
import tempfile
import time
from math import exp, inf
from pathlib import Path
from typing import Dict, List, Tuple, Optional

import numpy as np
from loguru import logger

from msserviceprofiler.modelevalstate.config.base_config import AnalyzeTool, BenchMarkPolicy, DeployPolicy
from msserviceprofiler.modelevalstate.config.base_config import CUSTOM_OUTPUT, custom_output, FOLDER_LIMIT_SIZE
from msserviceprofiler.modelevalstate.optimizer.utils import backup, remove_file, close_file_fp, get_folder_size
from msserviceprofiler.modelevalstate.optimizer.analyze_profiler import analyze as analyze_profiler
from msserviceprofiler.modelevalstate.common import get_train_sub_path, read_csv_s
from msserviceprofiler.msguard.security import open_s


_analyze_mapping = {AnalyzeTool.profiler.value: analyze_profiler}


def validate_parameters(common_generate_speed, perf_generate_token_speed, first_token_time, decode_time):
    if common_generate_speed is None and perf_generate_token_speed is None:
        raise ValueError("Not Found common_generate_speed or perf_generate_token_speed.")
    if first_token_time is None or decode_time is None:
        raise ValueError("Not Found first_token_time.")


class BenchMark:
    def __init__(self, benchmark_config, throughput_type: str = "common",
                 bak_path: Optional[Path] = None):
        from msserviceprofiler.modelevalstate.config.custom_command import BenchmarkCommand
        self.benchmark_config = benchmark_config
        self.throughput_type = throughput_type
        self.bak_path = bak_path
        self.run_log = None
        self.run_log_offset = None
        self.run_log_fp = None
        self.process = None
        self.pattern = re.compile(r'\(([^)]+)')
        self.command = BenchmarkCommand(self.benchmark_config.command).command

    def backup(self, del_log=True):
        backup(self.benchmark_config.output_path, self.bak_path, self.__class__.__name__)
        if not del_log:
            backup(self.run_log, self.bak_path, self.__class__.__name__)

    def get_performance_index(self):
        from msserviceprofiler.modelevalstate.config.config import PerformanceIndex

        output_path = Path(self.benchmark_config.output_path)
        common_generate_speed = None
        first_token_time = None
        perf_generate_token_speed = None
        decode_time = None
        success_rate = None
        for file in output_path.iterdir():
            if "result_common" in file.name:
                try:
                    df = read_csv_s(file)
                    if "OutputGenerateSpeed" in df.columns:
                        _generate_speed = df["OutputGenerateSpeed"][0]
                    else:
                        _generate_speed = df["GenerateSpeed"][0]
                    if isinstance(_generate_speed, str):
                        common_generate_speed = float(_generate_speed.split()[0])
                    elif isinstance(_generate_speed,
                                    (int, float, np.int64, np.int32, np.float64, np.float32, np.float16)):
                        common_generate_speed = _generate_speed
                    else:
                        raise TypeError(f"GenerateSpeed: {_generate_speed}, type: {type(_generate_speed)}")
                    req_returnd = df["Returned"][0]
                    if not req_returnd:
                        continue
                    _m_res = self.pattern.search(req_returnd)
                    if not _m_res:
                        continue
                    cleaned_str = _m_res.group(1).replace(' ', '').rstrip('%')
                    success_rate = float(cleaned_str) / 100
                except (KeyError, AttributeError) as e:
                    logger.error(f"Failed in get GenerateSpeed. error: {e}")
                continue
            if "result_perf" in file.name:
                try:
                    df = read_csv_s(file)
                    first_token_time = float(df["FirstTokenTime"][0].split()[0])
                    perf_generate_token_speed = float(df["GeneratedTokenSpeed"][0].split()[0])
                    decode_time = float(df["DecodeTime"][0].split()[0])
                except (AttributeError, KeyError) as e:
                    logger.error("Failed in get FirstTokenTime or GeneratedTokenSpeed. error: {}", e)
        validate_parameters(common_generate_speed, perf_generate_token_speed, first_token_time, decode_time)
        if self.throughput_type == "common":
            generate_speed = common_generate_speed
        else:
            generate_speed = perf_generate_token_speed
        time_to_first_token = first_token_time / 10 ** 3
        time_per_output_token = decode_time / 10 ** 3
        return PerformanceIndex(generate_speed=generate_speed, time_to_first_token=time_to_first_token,
                                time_per_output_token=time_per_output_token, success_rate=success_rate)

    def prepare(self):
        remove_file(Path(self.benchmark_config.output_path))
        remove_file(Path(self.benchmark_config.custom_collect_output_path))

    def check_success(self, print_log=False):
        if self.run_log:
            run_log_path = Path(self.run_log)
            if run_log_path.exists() and print_log:
                try:
                    with open_s(run_log_path, "r", encoding="utf-8") as f:
                        f.seek(self.run_log_offset)
                        output = f.read()
                        self.run_log_offset = f.tell()
                        logger.info(f"benchmark out: \n{output}")
                except (UnicodeError, OSError) as e:
                    logger.error(f"Failed read benchmark log. error {e}")
        try:
            if self.process.poll() is None:
                return False
            elif self.process.poll() == 0:
                return True
            else:
                raise subprocess.SubprocessError(
                    f"Failed in run benchmark. return code: {self.process.returncode}. ")
        except AttributeError as e:
            logger.error(f"Failed to check process status, error {e}")
            return False

    def run(self, run_params):
        # 启动测试
        logger.info("Start the benchmark test.")
        self.run_log_fp, self.run_log = tempfile.mkstemp(prefix="modelevalstate_benchmark")
        self.run_log_offset = 0
        if self.benchmark_config.work_path:
            cwd = self.benchmark_config.work_path
        else:
            cwd = os.getcwd()
        for k in run_params:
            if k.config_position == "env":
                try:
                    os.environ[k.name] = str(k.value)
                    _var_name = f"${k.name}"
                    if _var_name in self.command:
                        _i = self.command.index(_var_name)
                        self.command[_i] = str(k.value)
                except KeyError as e:
                    logger.error(f"Failed to set environment variable. error {e}")
        if CUSTOM_OUTPUT not in os.environ:
            os.environ[CUSTOM_OUTPUT] = str(custom_output)
        try:
            self.process = subprocess.Popen(self.command, env=os.environ, stdout=self.run_log_fp, 
                                            stderr=subprocess.STDOUT, text=True, cwd=cwd)
        except OSError as e:
            logger.error(f"Failed to run benchmark. error {e}")
            raise e
        logger.info(f"command: {' '.join(self.command)}, log file: {self.run_log}")

    def stop(self, del_log=True):
        self.backup(del_log)
        close_file_fp(self.run_log_fp)
        try:
            if self.process and self.process.poll() is None:
                self.process.kill()
        except AttributeError as e:
            logger.error(f"Failed to kill process. error {e}")
        if del_log:
            remove_file(Path(self.run_log))


class ProfilerBenchmark(BenchMark):
    def __init__(self, benchmark_config, *args, analyze_tool: AnalyzeTool = AnalyzeTool.default,
                 **kwargs):
        super().__init__(benchmark_config, *args, **kwargs)
        self.analyze_tool = analyze_tool
        self.profiler_cmd = ["python", "-m", "ms_service_profiler.parse",
                             f"--input-path={self.benchmark_config.profile_input_path}",
                             f"--output-path={self.benchmark_config.profile_output_path}"]
        self.profiler_log = None
        self.profiler_log_fp = None
        self.profiler_log_offset = 0
        self.profiler_process = None

    def extra_performance_index(self, *args, **kwargs):
        from msserviceprofiler.modelevalstate.config.config import PerformanceIndex

        logger.info("extra_performance_index")
        analyze_tool = _analyze_mapping.get(self.analyze_tool)
        if analyze_tool is None:
            raise ValueError(f"Analyze tool not found: {self.analyze_tool}")
        res = analyze_tool(*args, **kwargs)
        time_to_first_token = time_per_output_token = success_rate = None
        if isinstance(res, tuple):
            if len(res) == 1:
                generate_speed = res[0]
            elif len(res) == 2:
                generate_speed, time_to_first_token = res
            elif len(res) == 3:
                generate_speed, time_to_first_token, time_per_output_token = res
            elif len(res) == 4:
                generate_speed, time_to_first_token, time_per_output_token, success_rate = res
            else:
                raise ValueError(f"Not Support. res: {res}")
        else:
            generate_speed = res
        return PerformanceIndex(generate_speed=generate_speed, time_to_first_token=time_to_first_token,
                                time_per_output_token=time_per_output_token, success_rate=success_rate)

    def backup(self, del_log=True):
        super().backup(del_log)
        backup(self.benchmark_config.profile_input_path, self.bak_path, self.__class__.__name__)
        backup(self.benchmark_config.profile_output_path, self.bak_path, self.__class__.__name__)
        if not del_log and self.profiler_log:
            backup(self.profiler_log, self.bak_path, self.__class__.__name__)

    def prepare(self):
        super().prepare()
        remove_file(Path(self.benchmark_config.profile_input_path))
        remove_file(Path(self.benchmark_config.profile_output_path))

    def check_profiler(self, print_log=False):
        if print_log:
            try:
                with open_s(self.profiler_log, "r") as f:
                    f.seek(self.profiler_log_offset)
                    output = f.read()
                    self.profiler_log_offset = f.tell()
            except (UnicodeError, OSError) as e:
                logger.error(f"Failed read benchmark log. error {e}")
            if output:
                logger.info(f"benchmark out: \n{output}")
        if self.profiler_process.poll() is None:
            return False
        elif self.profiler_process.poll() == 0:
            return True
        else:
            raise subprocess.SubprocessError(
                f"Failed in run benchmark. return code: {self.process.returncode}. ")

    def start_profiler(self):
        self.profiler_log_fp, self.profiler_log = tempfile.mkstemp(prefix="modelevalstate_profiler")
        self.profiler_log_offset = 0
        if not os.path.exists(self.benchmark_config.work_path):
            raise FileNotFoundError(f"Work path not found: {self.benchmark_config.work_path}")
        logger.info(f"command: {' '.join(self.profiler_cmd)}, log file: {self.profiler_log}")
        self.profiler_process = subprocess.Popen(self.profiler_cmd, env=os.environ, stdout=self.profiler_log_fp,
                                                 stderr=subprocess.STDOUT,
                                                 text=True, cwd=self.benchmark_config.work_path)

    def get_performance_index(self):
        logger.info("get_performance_index")
        try:
            self.start_profiler()
            logger.info("wait profiler")
            while True:
                if self.check_profiler(print_log=True):
                    break
                time.sleep(1)
        except Exception as e:
            logger.error(f"Failed in start profiler. relation log: {self.profiler_log}")
            raise e
        collect_path = Path(self.benchmark_config.custom_collect_output_path)
        if self.analyze_tool == AnalyzeTool.profiler.value:
            res = self.extra_performance_index(self.benchmark_config.profile_output_path, collect_path)
            return res
        else:
            return super().get_performance_index()

    def stop(self, del_log=True):
        super().stop(del_log)
        close_file_fp(self.profiler_log_fp)
        if del_log:
            remove_file(Path(self.profiler_log))
        try:
            if self.profiler_process and self.profiler_process.poll() is None:
                self.profiler_process.kill()
        except AttributeError as e:
            logger.error(f"Failed to kill process. error {e}")


class VllmBenchMark(BenchMark):
    from msserviceprofiler.modelevalstate.config.custom_command import VllmBenchmarkCommand

    def __init__(self, benchmark_config, throughput_type: str = "common", bak_path: Optional[Path] = None):
        super().__init__(benchmark_config, throughput_type, bak_path)
        self.output_path = benchmark_config.output_path
        if not self.output_path.exists():
            self.output_path.mkdir(parents=True)
        self.command = VllmBenchmarkCommand(self.benchmark_config.vllm_command).command

    def get_performance_index(self):
        from msserviceprofiler.modelevalstate.config.config import PerformanceIndex

        output_path = Path(self.benchmark_config.output_path)
        generate_speed = None
        time_per_output_token = None
        time_to_first_token = None
        success_rate = None
        for file in output_path.iterdir():
            if not file.name.endswith(".json"):
                continue
            with open_s(file, mode='r', encoding="utf-8") as f:
                try:
                    data = json.load(f)
                except json.JSONDecodeError as err:
                    logger.warning("Failed to open %r, error: %r" % (file, err))
                    data = {}
            generate_speed = data.get("output_throughput", 0)
            time_to_first_token = data.get("mean_ttft_ms", 0) / 10 ** 3
            time_per_output_token = data.get("mean_tpot_ms", 0) / 10 ** 3
            num_prompts = data.get("num_prompts", 1)
            completed = data.get("completed", 0)
            if num_prompts != 0:
                success_rate = completed / num_prompts
        return PerformanceIndex(generate_speed=generate_speed,
                                time_to_first_token=time_to_first_token,
                                time_per_output_token=time_per_output_token,
                                success_rate=success_rate)

    def run(self, run_params):
        # 启动测试
        logger.info("Start the benchmark test.")
        self.run_log_fp, self.run_log = tempfile.mkstemp(prefix="modelevalstate_benchmark")
        self.run_log_offset = 0
        if self.benchmark_config.work_path:
            cwd = self.benchmark_config.work_path
        else:
            cwd = os.getcwd()
        for k in run_params:
            if k.config_position == "env":
                try:
                    os.environ[k.name] = str(k.value)
                    _var_name = f"${k.name}"
                    if _var_name in self.command:
                        _i = self.command.index(_var_name)
                        self.command[_i] = str(k.value)
                except KeyError as e:
                    logger.error(f"Failed to set environment variable. error {e}")
        if CUSTOM_OUTPUT not in os.environ:
            os.environ[CUSTOM_OUTPUT] = str(custom_output)
        os.environ["MODEL_EVAL_STATE_VLLM_CUSTOM_OUTPUT"] = str(self.output_path)
        _var_name = f"$MODEL_EVAL_STATE_VLLM_CUSTOM_OUTPUT"
        if _var_name in self.command:
            _i = self.command.index(_var_name)
            self.command[_i] = str(self.output_path)
        try:
            self.process = subprocess.Popen(self.command, env=os.environ, stdout=self.run_log_fp, 
                                            stderr=subprocess.STDOUT, text=True, cwd=cwd)
        except OSError as e:
            logger.error(f"Failed to run benchmark. error {e}")
            raise e
        logger.info(f"command: {' '.join(self.command)}, log file: {self.run_log}")


class Scheduler:
    def __init__(self, simulator, benchmark, data_storage,
                 bak_path: Optional[Path] = None, retry_number: int = 3, wait_start_time=1800):
        self.simulator = simulator
        self.benchmark = benchmark
        self.data_storage = data_storage
        self.bak_path = bak_path
        self.retry_number = retry_number
        self.wait_time = wait_start_time
        self.current_back_path = None
        self.simulate_run_info = None

    def back_up(self):
        if self.bak_path:
            if get_folder_size(self.bak_path) > FOLDER_LIMIT_SIZE:
                self.simulator.bak_path = None
                self.benchmark.bak_path = None
            else:
                _cur_bak_path = get_train_sub_path(self.bak_path)
                self.simulator.bak_path = _cur_bak_path
                self.benchmark.bak_path = _cur_bak_path

    def wait_simulate(self):
        logger.info("wait run simulator")
        for _ in range(self.wait_time):
            time.sleep(1)
            if self.simulator.check_success():
                logger.info(f"Successfully started the {self.simulator.process.pid} process.")
                return
        raise TimeoutError(self.wait_time)

    def run_simulate(self, params: np.ndarray, params_field):
        self.benchmark.prepare()
        self.simulator.run(tuple(self.simulate_run_info))
        self.wait_simulate()

    def monitoring_status(self):
        logger.info("monitor status")
        while True:
            if self.simulator.process.poll() is not None:
                self.simulator.stop(del_log=False)
                self.benchmark.stop(del_log=False)
                raise subprocess.SubprocessError(f"Failed in run simulator. "
                                                 f"return code: {self.simulator.process.returncode}.")
            if self.benchmark.check_success():
                return
            time.sleep(1)

    def run_target_server(self, params: np.ndarray, params_field):
        """
        1. 启动mindie仿真
        2. 启动benchmark 测试
        3. 检查mindie状态，检查benchmark状态
        """
        for _ in range(self.retry_number):
            try:
                self.run_simulate(params, params_field)
            except Exception as e:
                logger.error(f"Failed in Simulator Running. error: {e}， simulator log {self.simulator.mindie_log}")
                logger.exception("What?!")
                self.stop_target_server(del_log=False)
                continue
            time.sleep(1)
            try:
                self.benchmark.run(tuple(self.simulate_run_info))
            except Exception as e:
                logger.error(f"Failed in Benchmark Running. error: {e}, benchmark log {self.benchmark.run_log}")
                logger.exception("What?!")
                self.stop_target_server(del_log=False)
                continue
            time.sleep(1)
            try:
                self.monitoring_status()
            except Exception as e:
                self.stop_target_server(del_log=False)
                logger.error(f"Failed in monitoring status. error: {e}, simulator log {self.simulator.mindie_log}, "
                             f"benchmark log {self.benchmark.run_log}")
                logger.exception("What?!")
                continue
            return
        raise ValueError(
            f"Failed in run_target_server")

    def stop_target_server(self, del_log=True):
        self.simulator.stop(del_log)
        self.benchmark.stop(del_log)

    def run(self, params: np.ndarray, params_field):
        """
        1. 启动mindie仿真
        2. 启动benchmark 测试
        3. 获取benchmark测试结果
        4. 关闭mindie仿真
        5. 返回benchmark测试结果
        params: 是一维数组，其值对应mindie 的相关配置。
        """
        from msserviceprofiler.modelevalstate.config.config import map_param_with_value, PerformanceIndex

        logger.info("Start run in scheduler.")
        self.back_up()
        self.simulate_run_info = map_param_with_value(params, params_field)
        error_info = None
        del_log = True
        performance_index = PerformanceIndex()
        try:
            self.run_target_server(params, params_field)
            time.sleep(1)
            performance_index = self.benchmark.get_performance_index()
        except Exception as e:
            logger.error(f"Failed running. bak path: {self.simulator.bak_path}")
            error_info = e
            del_log = False
        self.data_storage.save(performance_index, tuple(self.simulate_run_info), error=error_info, 
                               backup=self.current_back_path)
        self.stop_target_server(del_log)
        if error_info:
            raise error_info
        return performance_index


class ScheduleWithMultiMachine(Scheduler):
    def __init__(self, communication_config, *args, **kwargs):
        from msserviceprofiler.modelevalstate.optimizer.communication import CommunicationForFile, CustomCommand

        super().__init__(*args, **kwargs)
        self.communication_config = communication_config
        self.communication = CommunicationForFile(self.communication_config.cmd_file,
                                                  self.communication_config.res_file)
        self.cmd = CustomCommand()
        _cmd = self.cmd.init
        self.communication.send_command(_cmd)
        self.communication.clear_command(_cmd)

    def back_up(self):
        if self.bak_path:
            if get_folder_size(self.bak_path) > FOLDER_LIMIT_SIZE:
                self.simulator.bak_path = None
                self.benchmark.bak_path = None
            else:
                _cur_bak_path = get_train_sub_path(self.bak_path)
                self.simulator.bak_path = _cur_bak_path
                self.benchmark.bak_path = _cur_bak_path
                _cmd = f"{self.cmd.backup} params:{_cur_bak_path}"
                self.communication.send_command(_cmd)
                self.communication.clear_cmd(_cmd)

    def monitoring_status(self):
        logger.info("Start monitoring")
        while True:
            _cmd = self.cmd.process_poll
            self.communication.send_command(_cmd)
            all_poll = [self.simulator.process.poll(), self.communication.clear_cmd(_cmd)]
            if any([_i is not None for _i in all_poll]):
                self.stop_target_server(del_log=False)
                raise subprocess.SubprocessError(
                    f"Failed in run simulator. all status: {all_poll}")
            if self.benchmark.check_success():
                return
            time.sleep(1)

    def run_simulate(self, params: np.ndarray, params_field):
        _cmd = f"{self.cmd.start} params:{params.tolist()}"
        self.cmd.history = _cmd
        self.communication.send_command(_cmd)
        self.communication.clear_cmd(_cmd)
        self.simulator.run(tuple(self.simulate_run_info))
        self.wait_simulate()
        # wait 其他服务器上服务成功
        _cmd = self.cmd.check_success
        self.cmd.history = _cmd
        self.communication.send_command(_cmd)
        self.communication.clear_cmd(_cmd)

    def stop_target_server(self, del_log=True):
        super(ScheduleWithMultiMachine, self).stop_target_server(del_log)
        # wait 其他服务器上服务成功
        _cmd = f"{self.cmd.stop} params:{del_log}"
        self.communication.send_command(_cmd)
        self.communication.clear_cmd(_cmd)
        self.cmd.history = _cmd


class PSOOptimizer:
    def __init__(self, scheduler: Scheduler, n_particles: int = 10, iters=100, pso_options=None,
                 target_field: Optional[Tuple] = None, prefill_lam: float = 0.5, decode_lam: float = 0.5,
                 success_rate_lam: float = 0.5, prefill_constraint: float = 0.05, decode_constraint: float = 0.05,
                 success_rate_constraint: float = 1, load_history_data: Optional[List] = None,
                 load_breakpoint: bool = False, pso_init_kwargs: Optional[Dict] = None):
        from msserviceprofiler.modelevalstate.config.config import default_support_field, PsoOptions

        self.scheduler = scheduler
        self.n_particles = n_particles
        self.iters = iters
        self.target_field = target_field if target_field else default_support_field
        if not pso_options:
            self.pso_options = PsoOptions()
        else:
            self.pso_options = pso_options
        self.prefill_lam = prefill_lam  # 优化算法中惩罚系数
        self.decode_lam = decode_lam
        self.success_rate_lam = success_rate_lam
        self.prefill_constraint = prefill_constraint
        self.decode_constraint = decode_constraint
        self.success_rate_constraint = success_rate_constraint
        self.load_history_data = load_history_data
        self.load_breakpoint = load_breakpoint
        self.pso_init_kwargs = pso_init_kwargs
        self.init_pos = None
        self.history_cost, self.history_pos = None, None
        if self.load_history_data and self.load_breakpoint:
            self.history_pos, self.history_cost = self.computer_fitness()

    def computer_fitness(self) -> Tuple:
        from msserviceprofiler.modelevalstate.config.config import PerformanceIndex

        all_position = []
        all_cost = []
        for case_data in self.load_history_data:
            _params = {}
            for k in PerformanceIndex.model_fields.keys():
                if k in case_data:
                    _params[k] = case_data[k]
            performance_index = PerformanceIndex(**_params)
            try:
                _fitness = self.minimum_algorithm(performance_index)
                logger.info(f"fitness {_fitness}")
                _pos = [case_data.get(_field.name) for _field in self.target_field]
                if not all(_pos):
                    continue
                all_cost.append(_fitness)
                all_position.append(_pos)
            except KeyError:
                continue
        if len(all_position) != len(all_cost):
            raise ValueError("Failed in computer_fitness.")
        return all_position, all_cost

    def minimum_algorithm(self, performance_index) -> float:
        if not isinstance(performance_index.generate_speed, (int, float)) or performance_index.generate_speed == 0:
            return inf
        try:
            fitness = 1 / performance_index.generate_speed
        except OverflowError:
            return inf
        
        def calculate_metric(value, constraint, lam):
            if constraint == 0:
                return inf
            try:
                _var = max(0.0, (value - constraint) / constraint)
                return lam * (exp(_var) - 1)
            except OverflowError:
                return inf
            
        if performance_index.time_to_first_token is not None:
            fitness += calculate_metric(performance_index.time_to_first_token, 
                                    self.prefill_constraint, 
                                    self.prefill_lam)
        if performance_index.time_per_output_token is not None:
            fitness += calculate_metric(performance_index.time_per_output_token,
                                      self.decode_constraint,
                                      self.decode_lam)
        if performance_index.success_rate:
            fitness += calculate_metric(self.success_rate_constraint - performance_index.success_rate,
                                      self.success_rate_constraint,
                                      self.success_rate_lam)
        return fitness

    def op_func(self, x) -> np.ndarray:
        n_particles = x.shape[0]
        logger.info(f"Acquired n_particles: {n_particles}, value: {x}")
        generate_speed = []
        for i in range(n_particles):
            # 调用schedule， 获取采集的数据
            try:
                _res = self.scheduler.run(x[i], self.target_field)
                # 根据采集的数据，计算最优化值
                _fitness = self.minimum_algorithm(_res)
            except Exception as e:
                logger.error(f"Failed. error: {e}, please check.")
                logger.exception("What?!")
                _fitness = inf
            logger.info(f"fitness {_fitness}")
            generate_speed.append(_fitness)
        return np.array(generate_speed)

    def constructing_bounds(self) -> Tuple[Tuple, Tuple]:
        """
        返回示例：((0, 10), (0, 10))
        """
        _min = []
        _max = []
        for _field in self.target_field:
            _min.append(_field.min)
            _max.append(_field.max)
        return (tuple(_min), tuple(_max))

    def run(self):
        from msserviceprofiler.modelevalstate.optimizer.global_best_custom import CustomGlobalBestPSO
        from msserviceprofiler.modelevalstate.config.config import map_param_with_value

        optimizer = CustomGlobalBestPSO(n_particles=self.n_particles, dimensions=len(self.target_field),
                                        options=self.pso_options.model_dump(), bounds=self.constructing_bounds(),
                                        init_pos=self.init_pos, breakpoint_pos=self.history_pos,
                                        breakpoint_cost=self.history_cost, **self.pso_init_kwargs)
        cost, joint_vars = optimizer.optimize(self.op_func, iters=self.iters)
        best_position = {_field.name: _field.value for _field in map_param_with_value(joint_vars, self.target_field)}
        logger.info(f"best cost {cost}, best joint_vars: {best_position}")


def arg_parse(subparsers):
    parser = subparsers.add_parser(
        "optimizer", formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="optimize for performance"
    )

    parser.add_argument("-lb", "--load_breakpoint", action="store_true",
                        help="Continue from where the last optimization was aborted.")
    parser.add_argument("-d", "--deploy_policy", default=DeployPolicy.single.value,
                        choices=[k.value for k in list(DeployPolicy)],
                        help="Indicates whether the multi-node running policy is used.")
    parser.add_argument("--backup", default=False, action="store_true",
                        help="Whether to back up data.")
    parser.add_argument("-b", "--benchmark_policy", default=BenchMarkPolicy.benchmark.value,
                        choices=[k.value for k in list(BenchMarkPolicy)],
                        help="Whether to use custom performance indicators.")
    parser.set_defaults(func=main)


def main(args: argparse.Namespace):
    from msserviceprofiler.modelevalstate.optimizer.server import main as slave_server
    from msserviceprofiler.modelevalstate.optimizer.simulator import Simulator, VllmSimulator
    from msserviceprofiler.modelevalstate.config.config import settings, ServiceType
    from msserviceprofiler.modelevalstate.optimizer.store import DataStorage

    if settings.service == ServiceType.slave.value:
        slave_server()
        return
    if args.benchmark_policy == BenchMarkPolicy.vllm_benchmark.value:
        simulator = VllmSimulator(settings.simulator)
    else:
        simulator = Simulator(settings.simulator)
    bak_path = None
    if args.backup:
        bak_path = settings.output.joinpath("bak")
        if not bak_path.exists():
            bak_path.mkdir(parents=True)
    # 单机benchmark
    if args.benchmark_policy == BenchMarkPolicy.benchmark.value:
        benchmark = BenchMark(settings.benchmark, bak_path=bak_path)
    elif args.benchmark_policy == BenchMarkPolicy.vllm_benchmark.value:
        benchmark = VllmBenchMark(settings.benchmark, bak_path=bak_path)
    elif args.benchmark_policy == BenchMarkPolicy.profiler_benchmark:
        benchmark = ProfilerBenchmark(settings.benchmark, bak_path=bak_path, analyze_tool=AnalyzeTool.profiler)
    else:
        benchmark = BenchMark(settings.benchmark, bak_path=bak_path)

    # 存储结果，只在主节点存储结果
    data_storage = DataStorage(settings.data_storage)
    # 初始化调度模块，支持单机和多机。
    if args.deploy_policy == DeployPolicy.multiple.value:
        scheduler = ScheduleWithMultiMachine(settings.communication, simulator, benchmark, data_storage,
                                             bak_path=bak_path)
    else:
        scheduler = Scheduler(simulator, benchmark, data_storage, bak_path=bak_path)
    _load_history_data = None
    _load_history = None
    if args.load_breakpoint:
        _load_history = True
    if _load_history:
        _load_history_data = data_storage.load_history_position(settings.data_storage.store_dir)
    pso = PSOOptimizer(scheduler, n_particles=settings.n_particles, iters=settings.iters,
                       prefill_lam=settings.prefill_lam, target_field=settings.target_field,
                       decode_lam=settings.decode_lam, success_rate_lam=settings.success_rate_lam,
                       decode_constraint=settings.decode_constraint, prefill_constraint=settings.prefill_constraint,
                       success_rate_constraint=settings.success_rate_constraint, load_history_data=_load_history_data,
                       load_breakpoint=args.load_breakpoint,
                       pso_init_kwargs={"ftol": settings.ftol, "ftol_iter": settings.ftol_iter})
    pso.run()



# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.
import os
import ast
import atexit
from copy import deepcopy
from pathlib import Path
from typing import Any, List, Tuple, Type, Optional, Union

import numpy as np
from loguru import logger
from pydantic import BaseModel, field_validator, Field
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource, JsonConfigSettingsSource
from msserviceprofiler.modelevalstate.config.custom_command import BenchmarkCommandConfig, VllmBenchmarkCommandConfig
from msserviceprofiler.modelevalstate.config.custom_command import MindieCommandConfig, VllmCommandConfig
from .base_config import INSTALL_PATH, RUN_PATH, ServiceType, custom_output, CUSTOM_OUTPUT, DeployPolicy, RUN_TIME
from .base_config import modelevalstate_config_path, MODEL_EVAL_STATE_CONFIG_PATH, AnalyzeTool, BenchMarkPolicy


class OptimizerConfigField(BaseModel):
    name: str = "max_batch_size"
    config_position: str = "BackendConfig.ScheduleConfig.maxBatchSize"
    min: int = 0
    max: int = 100
    dtype: str = "float"
    value: Union[int, float, bool] = 0.0
    dtype_param: Any = None


def map_param_with_value(params: np.ndarray, params_field: Tuple[OptimizerConfigField]):
    _simulate_run_info = []
    for i, v in enumerate(params_field):
        _field = deepcopy(v)
        if v.dtype == "int":
            _field.value = int(params[i])
        elif v.dtype == "bool":
            if params[i] > 0.5:
                _field.value = True
            else:
                _field.value = False
        elif v.dtype == "enum":
            segment = np.linspace(v.min, v.max, len(v.dtype_param) + 1)
            if params[i] <= v.min:
                _field.value = v.dtype_param[0]
            elif params[i] >= v.max:
                _field.value = v.dtype_param[-1]
            else:
                _enum_index = np.searchsorted(segment, params[i]) - 1
                _field.value = v.dtype_param[_enum_index]
        else:
            _field.value = float(params[i])
        _simulate_run_info.append(_field)
    for i, v in enumerate(params_field):
        _field = _simulate_run_info[i]
        if v.dtype == "ratio":
            _t_op = [_op for _op in _simulate_run_info if _op.name == v.dtype_param][0]
            _field.value = int(_field.value * _t_op.value)
        if v.dtype == "factories":
            _t_op = [_op for _op in _simulate_run_info if _op.name == v.dtype_param["target_name"]][0]
            if _t_op.value != 0:
                _field.value = ast.literal_eval(v.dtype_param["dtype"])(v.dtype_param["product"] / _t_op.value)
        if "maxPrefillBatchsize" in v.config_position:
            _field = _simulate_run_info[i]
            if _field.value == 0:
                _field.value = 1

    return _simulate_run_info


class PerformanceIndex(BaseModel):
    generate_speed: Optional[float] = None
    time_to_first_token: Optional[float] = None
    time_per_output_token: Optional[float] = None
    success_rate: Optional[float] = None


class BenchMarkConfig(BaseModel):
    name: str = "benchmark"
    work_path: Path = Field(default_factory=lambda: Path(os.getcwd()).resolve())
    command: BenchmarkCommandConfig = BenchmarkCommandConfig()
    vllm_command: VllmBenchmarkCommandConfig = VllmBenchmarkCommandConfig()
    output_path: Path = custom_output.joinpath("instance")
    custom_collect_output_path: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/custom_collect_output_path").resolve(),
        validate_default=True
    )
    profile_input_path: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/profile_input_path").resolve(),
        validate_default=True
    )
    profile_output_path: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/profile_output_path").resolve(),
        validate_default=True
    )

    @field_validator("output_path", "custom_collect_output_path", "profile_input_path",
                     "profile_output_path")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path

    @field_validator("work_path")
    @classmethod
    def check_dir(cls, path: Path) -> Path:
        if not path.exists():
            logger.error(f"FileNotFound: {path!r}")
        return path


class CommunicationConfig(BaseModel):
    cmd_file: Optional[Path] = custom_output.joinpath("cmd.txt")
    res_file: Optional[Path] = custom_output.joinpath("res.txt")


class DataStorageConfig(BaseModel):
    store_dir: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/store").resolve(),
        validate_default=True
    )

    @field_validator("store_dir")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path
    

class LatencyModel(BaseModel):
    base_path: Path = Field(default_factory=lambda: custom_output.joinpath("result/latency_model").resolve())
    model_path: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("bak/base/xgb_model.ubj").resolve())
    ohe_path: Optional[Path] = Field(default_factory=lambda data: data["base_path"].joinpath("ohe").resolve())
    static_file_dir: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("model_static_file").resolve(), validate_default=True)
    req_and_decode_file: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("req_id_and_decode_num.json").resolve())
    cache_data: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("cache").resolve())

    @field_validator("static_file_dir")
    @classmethod
    def validate_cache_data(cls, path: Optional[Path] = None):
        if path:
            if not Path(path).exists():
                path.mkdir(parents=True)
        return path
    
    @field_validator("static_file_dir")
    @classmethod
    def validate_static_file_dir(cls, path: Optional[None]):
        if path is None:
            path = cls.base_path.joinpath("model_static_file")
            path.mkdir(parents=True)
        return path


class MindieConfig(BaseModel):
    # 运行mindie时，要修改的mindie config
    mindie_service_default_path: str = "/usr/local/Ascend/mindie/latest/mindie-service"
    mindie_service_path: str = os.getenv("MIES_INSTALL_PATH", mindie_service_default_path)
    process_name: str = "mindieservice_daemon"
    config_path: Path = Path(os.path.join(mindie_service_path, "conf", "config.json"))
    config_bak_path: Path = Path(os.path.join(mindie_service_path, "conf", "config_bak.json"))
    work_path: Path = Field(default_factory=lambda: Path(os.getcwd()).resolve())
    command: MindieCommandConfig = MindieCommandConfig()
    vllm_command: VllmCommandConfig = VllmCommandConfig()
    log_path: Path = Path(os.path.join(mindie_service_path, "logs"))


    @field_validator("config_path")
    @classmethod
    def check_dir(cls, path: Path) -> Path:
        if not path.exists():
            logger.error(f"FileNotFound: {path!r}")
        return path

    @field_validator("log_path")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path


@atexit.register
def clearing_residual_process():
    from ..optimizer.utils import kill_process

    kill_process(MindieConfig().process_name)


class PsoOptions(BaseModel):
    c1: float = 2.0
    c2: float = 2.0
    w: float = 1.8


class PsoStrategy(BaseModel):
    w: str = "exp_decay"
    c1: str = "exp_decay"
    c2: str = "exp_decay"


default_support_field = [
    # max batch size 最小值要大于max_prefill_batch_size的最大值。
    OptimizerConfigField(name="max_batch_size", config_position="BackendConfig.ScheduleConfig.maxBatchSize", min=25,
                         max=300, dtype="int"),
    OptimizerConfigField(name="max_prefill_batch_size",
                         config_position="BackendConfig.ScheduleConfig.maxPrefillBatchSize", min=1, max=25,
                         dtype="int"),
    OptimizerConfigField(name="prefill_time_ms_per_req",
                         config_position="BackendConfig.ScheduleConfig.prefillTimeMsPerReq", max=1000, dtype="int"),
    OptimizerConfigField(name="decode_time_ms_per_req",
                         config_position="BackendConfig.ScheduleConfig.decodeTimeMsPerReq", max=1000, dtype="int"),
    OptimizerConfigField(name="support_select_batch",
                         config_position="BackendConfig.ScheduleConfig.supportSelectBatch", max=1,
                         dtype="bool"),
    OptimizerConfigField(name="max_prefill_token",
                         config_position="BackendConfig.ScheduleConfig.maxPrefillTokens", min=4096, max=409600,
                         dtype="int"),
    OptimizerConfigField(name="max_queue_deloy_mircroseconds",
                         config_position="BackendConfig.ScheduleConfig.maxQueueDelayMicroseconds", min=500, max=1000000,
                         dtype="int"),
    OptimizerConfigField(name="prefill_policy_type",
                         config_position="BackendConfig.ScheduleConfig.prefillPolicyType", min=0, max=1,
                         dtype="enum", dtype_param=[0, 1, 3]),
    OptimizerConfigField(name="decode_policy_type",
                         config_position="BackendConfig.ScheduleConfig.decodePolicyType", min=0, max=1,
                         dtype="enum", dtype_param=[0, 1, 3]),
    OptimizerConfigField(name="max_preempt_count",
                         config_position="BackendConfig.ScheduleConfig.maxPreemptCount", min=0, max=1,
                         dtype="ratio", dtype_param="max_batch_size")
]


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        json_file=[INSTALL_PATH.joinpath("model_eval_state.json"), Path("~/model_eval_state.json").expanduser(),
                   RUN_PATH.joinpath("model_eval_state.json"),
                   INSTALL_PATH.joinpath("config.json"), Path("~/config.json").expanduser(),
                   RUN_PATH.joinpath("config.json"), modelevalstate_config_path],
        env_prefix="model_eval_state_")
    output: Path = Field(default_factory=lambda: custom_output.joinpath("result").resolve(), validate_default=True)
    latency_model: LatencyModel = LatencyModel()
    simulator: MindieConfig = MindieConfig()
    benchmark: BenchMarkConfig = BenchMarkConfig()
    data_storage: DataStorageConfig = DataStorageConfig()
    pso_options: PsoOptions = PsoOptions()
    pso_strategy: PsoStrategy = PsoStrategy()
    n_particles: int = 5
    iters: int = 10
    ftol: float = -np.inf
    ftol_iter: int = 1
    prefill_lam: float = 0.5  # 惩罚系数
    decode_lam: float = 0.5
    success_rate_lam: float = 0.5
    prefill_constraint: float = 0.05
    decode_constraint: float = 0.05
    success_rate_constraint: float = 1.0
    service: str = ServiceType.master
    communication: CommunicationConfig = CommunicationConfig()
    float_range_in_best_particle: float = 0.1  # 如果用历史值作为作为初始值，那么允许在初始值的浮动程度。
    target_field: List[OptimizerConfigField] = default_support_field

    @classmethod
    def settings_customise_sources(
            cls,
            settings_cls: Type[BaseSettings],
            init_settings: PydanticBaseSettingsSource,
            env_settings: PydanticBaseSettingsSource,
            dotenv_settings: PydanticBaseSettingsSource,
            file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (init_settings, env_settings, JsonConfigSettingsSource(settings_cls), file_secret_settings)

    @field_validator("output")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path


settings = Settings()
